//---------------------------------------------------------------------------
// VTreeHolder.cpp
// (c) RVA (r Doga, or BARDAKC) (c)
//    TVirtualStringTree  IShellFolder
//---------------------------------------------------------------------------
#pragma hdrstop
//---------------------------------------------------------------------------
#include "VTreeHolder.h"
#include <FileUtil.hpp>
#include <winreg.h>
#include "AppUtilites.h"
//---------------------------------------------------------------------------
// TNodeData
//---------------------------------------------------------------------------
__fastcall TNodeData::TNodeData(void) 
{
  FNodeLevel   = 0;
  FNodeType    = ntNone;
  FNodeStatus  = nsOK;
  FHasChildren = false;
  FFullPath    = "";
  FNodeText    = "";
  FNodeHint    = "";
  FPIDL        = NULL; 
  FShellFolder = NULL; 
  FShellMallocOwner = NULL;
}
//---------------------------------------------------------------------------
__fastcall TNodeData::~TNodeData()
{
  if ((FNodeType == ntLocalDrive) || (FNodeType == ntNetDrive))
  {     
                          
    if (FPIDL)
    {
      if (FShellMallocOwner)
      {
        if (FShellMallocOwner->FpShellMalloc)
        {
          FShellMallocOwner->FpShellMalloc->Free(FPIDL);
        }  
      }  
    }

    if (FFullPIDL)
    {
      if (FShellMallocOwner)
      {
        if (FShellMallocOwner->FpShellMalloc)
        {
          FShellMallocOwner->FpShellMalloc->Free(FFullPIDL);
        }  
      }  
    }
    
    if (FShellFolder)
    {
      FShellFolder->Release();
    }
  }  
}
//---------------------------------------------------------------------------
void __fastcall TNodeData::Init(int ANodeLevel, 
                                NT ANodeType, 
                                NS ANodeStatus, 
                                bool AHasChildren, 
                                int ANodeClosedImageIndex, 
                                int ANodeOpenImageIndex, 
                                AnsiString  AFullPath, 
                                AnsiString ANodeText, 
                                AnsiString  ANodeHint, 
                                TVTreeHolderOperator *AShellMallocOwner, 
                                TItemIDList *APIDL, 
                                TItemIDList *AFullPIDL, 
                                IShellFolder *AShellFolder)
{
  FNodeLevel   = ANodeLevel;
  FNodeType    = ANodeType;
  FNodeStatus  = ANodeStatus;
  FHasChildren = AHasChildren;
  FNodeClosedImageIndex = ANodeClosedImageIndex;
  FNodeOpenImageIndex    = ANodeOpenImageIndex;
  FFullPath    = AFullPath;
  FNodeText    = ANodeText;
  FNodeHint    = ANodeHint;
  
  FPIDL = APIDL;
  FFullPIDL = AFullPIDL;
  FShellFolder = AShellFolder;
  FShellMallocOwner = AShellMallocOwner;
}
//---------------------------------------------------------------------------
// TVTreeHolderOperator
//---------------------------------------------------------------------------
__fastcall TVTreeHolderOperator::TVTreeHolderOperator(void)
{
  FpShellMalloc           = NULL;
  FVTreeViewHolder        = NULL;
  FHolderImages           = NULL;

  FImageListConstCount = 0;
}
//---------------------------------------------------------------------------
__fastcall TVTreeHolderOperator::~TVTreeHolderOperator()
{
  if (FHolderImages)
  {
    for (int k = FImageListConstCount; k > FHolderImages->Count;)
    {
      FHolderImages->Delete(k);
    }
  }

  if (FpShellMalloc)
  {  
    FpShellMalloc->Release();
  }
    
  CoUninitialize();
}
//---------------------------------------------------------------------------
void __fastcall TVTreeHolderOperator::Init(TVirtualStringTree  *AVTreeViewHolder, TImageList *AHolderImages, /*IGlobal *AGlobal,*/ MM AMonitorMode)
{
  FVTreeViewHolder = AVTreeViewHolder;
  FHolderImages    = AHolderImages;
  //FGlobal          = AGlobal;
  FMonitorMode     = AMonitorMode;

  if (FVTreeViewHolder && FHolderImages)
  {
    // 
    FVTreeViewHolder->RootNodeCount = 0;
    FVTreeViewHolder->NodeDataSize  = sizeof(VTVNODEDATA);

    FVTreeViewHolder->OnFreeNode        = NodeFreeEvent;
    FVTreeViewHolder->OnCollapsed       = NodeCollapsedEvent;
    FVTreeViewHolder->OnExpanding       = NodeExpandingEvent;
    FVTreeViewHolder->OnGetImageIndex   = NodeGetImageIndexEvent;
    FVTreeViewHolder->OnGetText         = NodeGetTextEvent;
    FVTreeViewHolder->OnPaintText = NodeBeforePaintEvent;
    
    FVTreeViewHolder->Images = AHolderImages;

    FImageListConstCount = FHolderImages->Count;

    TVirtualNode *TmpNode;
    VTVNODEDATA  *VTVNodeData;
    TNodeData    *FCurrentCreatedNodeData; 

    NS           NodeStatus;

    bool         HasChildren;

    
    //////////////////////////
    //  
    //////////////////////////
    		
    // 
    NodeStatus = nsDisabled /* ! */;
    HasChildren = false;
    FCurrentCreatedNodeData = new TNodeData();
    FCurrentCreatedNodeData->Init(0, ntFavorite, NodeStatus, HasChildren, 14, 14, "", "", "   \"\"", NULL, NULL, NULL, NULL) ;

    TmpNode = FVTreeViewHolder->AddChild(NULL); 

    VTVNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
    VTVNodeData->NodeData = FCurrentCreatedNodeData;

    if (FCurrentCreatedNodeData->FHasChildren)
    {
      TmpNode->States << vsHasChildren;
    } 
                
    //    
    FHResult = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED );
    if (SUCCEEDED(FHResult))
    {
      try
      {
        FHResult = SHGetMalloc(&FpShellMalloc);
        if (SUCCEEDED(FHResult))
        {
          try
          {
            IShellFolder *DesktopFolder = NULL;
            FHResult = SHGetDesktopFolder(&DesktopFolder);
            if (SUCCEEDED(FHResult))
            {
              try
              {
                TItemIDList  *TmpPIDL = NULL;
                IShellFolder *TmpShellFolder = NULL;  //???
                STRRET       StrRetValue;
          
                // 
                AnsiString HolderDirectory = DataPath;    //FGlobal->Parameter["FileHolderDirectory"];

                FCurrentCreatedNodeData = new TNodeData();
                
                if (DirectoryExists(HolderDirectory))
                {   
                  if (FMonitorMode == mmDAT)
                  {
                    HolderDirectory = NormalDir(HolderDirectory) + "DAT";
                  }
                  else //if (FMonitorMode == mmRXD)
                  {
                    HolderDirectory = NormalDir(HolderDirectory) + "RXD";
                  }
                  
                  wchar_t HolderDirectoryWCStr[MAX_PATH]; 
                  HolderDirectory.WideChar(HolderDirectoryWCStr, HolderDirectory.WideCharBufSize());
                
                  FHResult = DesktopFolder->ParseDisplayName(NULL, NULL, HolderDirectoryWCStr, NULL, &TmpPIDL, NULL);
                  if (SUCCEEDED(FHResult))
                  {
                    FHResult = DesktopFolder->BindToObject(TmpPIDL, NULL, IID_IShellFolder, (LPVOID*)&TmpShellFolder);
                    if (SUCCEEDED(FHResult))
                    {
                      if (!GetHasSubItems(TmpShellFolder, SHCONTF_FOLDERS))
                      {
                        HasChildren = false;
                      }
                      else
                      {
                        HasChildren = true;
                      }
                      
                      NodeStatus = nsOK;
                      FCurrentCreatedNodeData->Init(0, ntDesnaHolder, NodeStatus, HasChildren, 0, 0, "", "", "   \"\"", this, TmpPIDL, TmpPIDL, TmpShellFolder);
                      
                      goto HOLDER_IS_OK;
                    }
                  }
                }

                HasChildren = false;
                FCurrentCreatedNodeData->Init(0, ntDesnaHolder, nsDisabled, HasChildren, 0, 0, HolderDirectory, "", "   \"\"", NULL, NULL, NULL, NULL);

HOLDER_IS_OK:    

                TmpNode = FVTreeViewHolder->AddChild(NULL); 

                VTVNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
                VTVNodeData->NodeData = FCurrentCreatedNodeData;

                if (FCurrentCreatedNodeData->FHasChildren)
                {
                  TmpNode->States << vsHasChildren;
                }

                
                //   ( )
                FHResult = SHGetSpecialFolderLocation(NULL, CSIDL_DRIVES, &TmpPIDL);
                if (SUCCEEDED(FHResult))
                {
                  FHResult = DesktopFolder->BindToObject(TmpPIDL, NULL, IID_IShellFolder, (LPVOID*)&TmpShellFolder);
                  if (SUCCEEDED(FHResult))
                  {
                    FHResult = DesktopFolder->GetDisplayNameOf(TmpPIDL, SHGDN_INFOLDER, &StrRetValue);
                    if (SUCCEEDED(FHResult))
                    {
                      // 
                      char * TmpDisplayName = GetDisplayName(TmpPIDL, StrRetValue);

                      //   
                      int IconIndex;

                      TIcon *Icon = new TIcon(); 
                      
                      //    
                      Icon->Handle     = (void *)GetShellImage(TmpPIDL, false);
                      IconIndex = FHolderImages->AddIcon(Icon);
                      Icon->ReleaseHandle();
                      delete Icon;


                      FCurrentCreatedNodeData = new TNodeData();

                      if (!GetHasSubItems(TmpShellFolder, SHCONTF_FOLDERS))
                      {
                        //NodeStatus = nsDisabled;
                        HasChildren = false;
                      }
                      else
                      {
                        //NodeStatus = nsOK;
                        HasChildren = true;
                      }
                      
                      NodeStatus = nsOK;
                      FCurrentCreatedNodeData->Init(0, ntLocalDrive, NodeStatus, HasChildren, IconIndex, IconIndex, "", AnsiString(TmpDisplayName), "", this, TmpPIDL, TmpPIDL, TmpShellFolder);
                      free(reinterpret_cast<void *>(TmpDisplayName));

                      TmpNode = FVTreeViewHolder->AddChild(NULL); 

                      VTVNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
                      VTVNodeData->NodeData = FCurrentCreatedNodeData;

                      if (FCurrentCreatedNodeData->FHasChildren)
                      {
                        TmpNode->States << vsHasChildren;
                      }
                    }
                  }
                }

                //
                FHResult = SHGetSpecialFolderLocation(NULL, CSIDL_NETWORK, &TmpPIDL);
                if (SUCCEEDED(FHResult))
                {
                  FHResult = DesktopFolder->BindToObject(TmpPIDL, NULL, IID_IShellFolder, (LPVOID*)&TmpShellFolder);
                  if (SUCCEEDED(FHResult))
                  {
                    FHResult = DesktopFolder->GetDisplayNameOf(TmpPIDL, SHGDN_INFOLDER, &StrRetValue);
                    if (SUCCEEDED(FHResult))
                    {
                      // 
                      char * TmpDisplayName = GetDisplayName(TmpPIDL, StrRetValue);

                      //   
                      int IconIndex;

                      TIcon *Icon = new TIcon(); 
                      
                      //   
                      Icon->Handle     = (void *)GetShellImage(TmpPIDL, false);
                      IconIndex = FHolderImages->AddIcon(Icon);
                      Icon->ReleaseHandle();
                      delete Icon;

                      FCurrentCreatedNodeData = new TNodeData();

                      if (!GetHasSubItems(TmpShellFolder, SHCONTF_FOLDERS))
                      {
                        HasChildren = false;
                      }
                      else
                      {
                        HasChildren = true;
                      }
                      
                      NodeStatus = nsOK;
                      FCurrentCreatedNodeData->Init(0, ntNetDrive, NodeStatus, HasChildren, IconIndex, IconIndex, "", AnsiString(TmpDisplayName), "", this, TmpPIDL, TmpPIDL, TmpShellFolder);
                      free(reinterpret_cast<void *>(TmpDisplayName));

                      TmpNode = FVTreeViewHolder->AddChild(NULL); 

                      VTVNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
                      VTVNodeData->NodeData = FCurrentCreatedNodeData;

                      if (FCurrentCreatedNodeData->FHasChildren)
                      {
                        TmpNode->States << vsHasChildren;
                      }
                    }
                  }
                }
              }
              __finally
              {
                if (DesktopFolder)
                {
                  DesktopFolder->Release();
                }
              }
            }  
          }
          catch(...)
          {
            if (FpShellMalloc)
            {  
              FpShellMalloc->Release();
            }
            FpShellMalloc = NULL;
          }
        }  
      }
      catch(...)
      {
        CoUninitialize();
      }
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TVTreeHolderOperator::NodeFreeEvent(TBaseVirtualTree *Sender, PVirtualNode Node)
{
  VTVNODEDATA *VTVNodeData = (VTVNODEDATA *)Sender->GetNodeData(Node); 
  if (VTVNodeData)
  {
    delete VTVNodeData->NodeData;
  }  
}
//---------------------------------------------------------------------------
void __fastcall TVTreeHolderOperator::NodeCollapsedEvent(TBaseVirtualTree *Sender, PVirtualNode Node)
{
   //" - MSN"
}
//---------------------------------------------------------------------------
void __fastcall TVTreeHolderOperator::NodeExpandingEvent(TBaseVirtualTree *Sender, PVirtualNode Node, 
  bool &Allowed)
{
  TNodeData   *FCurrentCreatedNodeData;

  VTVNODEDATA *VTVNodeData = (VTVNODEDATA *)Sender->GetNodeData(Node); 

  if (VTVNodeData->NodeData)
  {
    NT   NodeType = VTVNodeData->NodeData->FNodeType;

    //   
    if (NodeType == ntFavorite)
    {
      return;
    }
    
    
    VTVNODEDATA *VTVChildNodeData;
    NT   ChildNodeType;
    bool ChildHasChildren;
    NS   ChildNodeStatus;

    //   
    int  ChildClosedIconIndex;
    int  ChildOpenIconIndex;

    LPENUMIDLIST Iterator;

    TVirtualNode *TmpNode;

    //    " - MSN"   
    AnsiString WebMSN = "";
    if (VTVNodeData->NodeData->FNodeLevel == 0)
    {
      if (VTVNodeData->NodeData->FNodeType == ntNetDrive)
      {
        HKEY  hTemp;  
                    
        FHResult = RegOpenKeyEx(HKEY_CURRENT_USER,
                                TEXT("Software\\Microsoft\\Web Service Providers\\WebDrive\\www.msnusers.com\\"),
                                0,
                                KEY_READ,
                                &hTemp);
                                            
        if (SUCCEEDED(FHResult))
        { 
          if (hTemp)
          {
            DWORD wType;
            DWORD bData = MAX_PATH;
            char RetValue[MAX_PATH];
            ZeroMemory(RetValue, MAX_PATH);
                      
            FHResult = RegQueryValueEx(hTemp,
                                       TEXT("Shortcut"),
                                       NULL,
                                       &wType,
                                       RetValue,
                                       &bData);

            if (SUCCEEDED(FHResult))
            {                           
              WebMSN = AnsiString(RetValue);
            }
          }
        }
        RegCloseKey(hTemp);
      }
    }             
              
    FHResult = VTVNodeData->NodeData->FShellFolder->EnumObjects(NULL, SHCONTF_FOLDERS/* | SHCONTF_INCLUDEHIDDEN*/, &Iterator);
    if (SUCCEEDED(FHResult))
    {
      Screen->Cursor = crHourGlass;
      
      try
      {
        TStringList *PlantList  = NULL;
        TStringList *UnitList   = NULL;
        TStringList *RepairList = NULL;

        if (NodeType == ntDesnaHolder)
        {
          /*
          PlantList = (FGlobal->GetSQLFunctions())->GetPlantsList();
          if (PlantList)
          {
            for (int k = 0; k < PlantList->Count; k++)
            {
              PlantList->Strings[k] = PlantList->Strings[k] + "=0";
            }
          }
          */
          PlantList = new TStringList();  
          PlantList->Text = " =0\n =0\n =0";
        }
        else if (NodeType == ntDesnaHolderPlant)
        {
          /*
          UnitList = (FGlobal->GetSQLFunctions())->GetUnitsList(VTVNodeData->NodeData->FNodeText);
          if (UnitList)
          {
            for (int k = 0; k < UnitList->Count; k++)
            {
              UnitList->Strings[k] = AnsiString(" ") + UnitList->Strings[k] +  AnsiString("=0");
            }
          }
          */  

          UnitList =  new TStringList();  
          UnitList->Text = " 1=0\n 2=0\n 3=0";
          
        }
        else if (NodeType == ntDesnaHolderUnit)
        {
          VTVNODEDATA *VTVParentNodeData = (VTVNODEDATA *)Sender->GetNodeData(Node->Parent); 
        
          AnsiString Plant = VTVParentNodeData->NodeData->FNodeText;
          AnsiString Unit  = VTVNodeData->NodeData->FNodeText;
          Unit             = Unit.Delete(1, 5);

          /*        
          RepairList = (FGlobal->GetSQLFunctions())->GetRepairsList(Plant, Unit);
          if (RepairList)
          {
            for (int k = 0; k < RepairList->Count; k++)
            {
              RepairList->Strings[k] = StringReplace(RepairList->Strings[k], "=", "-", TReplaceFlags()<<rfReplaceAll) + "=0";
            }
          } 
          */ 

          RepairList = new TStringList();  
          RepairList->Text = "22.04.2001-07.11.2001=0\n25.03.2003-07.08.2003=0\n04.04.2004-26.05.2004=0\n07.09.1999-14.10.1999=0\n08.06.2000-20.10.2000=0\n31.05.2003-26.07.2003=0";
        }
      
        for(;;)
        {
          IShellFolder *TmpShellFolder = NULL; 
              
          ULONG celtFetched;
          TItemIDList *TmpPIDL;
          TItemIDList *TmpFullPIDL;
          STRRET      StrRetValue;
            
          FHResult = Iterator->Next(0, &TmpPIDL, &celtFetched);

          if(FHResult != NOERROR) 
          {
            break;
          }  

          try
          {
            FHResult = VTVNodeData->NodeData->FShellFolder->BindToObject(TmpPIDL, NULL, IID_IShellFolder, (LPVOID*)&TmpShellFolder);
            if (SUCCEEDED(FHResult))
            {
              FHResult = VTVNodeData->NodeData->FShellFolder->GetDisplayNameOf(TmpPIDL, SHGDN_INFOLDER, &StrRetValue);
              if (SUCCEEDED(FHResult))
              {
                TmpFullPIDL = ConcatPIDLs(VTVNodeData->NodeData->FFullPIDL, TmpPIDL);
                
                // 
                char * TmpDisplayName1;
                char * TmpDisplayName2;

                bool PidlIsGood = true;

                TmpDisplayName1 = GetDisplayName(TmpPIDL, StrRetValue);

                FHResult = VTVNodeData->NodeData->FShellFolder->GetDisplayNameOf(TmpPIDL, SHGDN_NORMAL | SHGDN_FORPARSING, &StrRetValue);
                if (SUCCEEDED(FHResult))
                {
                  TmpDisplayName2 = GetDisplayName(TmpPIDL, StrRetValue);
                }
                else
                {
                  TmpDisplayName2 = "";
                }

                //...
                if ( ((VTVNodeData->NodeData->FNodeLevel == 0) && (VTVNodeData->NodeData->FNodeType == ntLocalDrive))
                  || ((VTVNodeData->NodeData->FNodeLevel >  0) && (VTVNodeData->NodeData->FNodeType == ntNetDrive)) ) 
                {
                  
                  // : 
                  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                  //HKEY_CLASSES_ROOT\CLSID\{21EC2020-3AEA-1069-A2DD-08002B30309D}  
                  //HKEY_CLASSES_ROOT\CLSID\{2227A280-3AEA-1069-A2DE-08002B30309D} 
                  //HKEY_CLASSES_ROOT\CLSID\{D6277990-4C6A-11CF-8D87-00AA0060F5BF}  
                  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                  
                  if ( (AnsiStrPos(TmpDisplayName2, "{21EC2020-3AEA-1069-A2DD-08002B30309D}") != NULL)
                    || (AnsiStrPos(TmpDisplayName2, "{2227A280-3AEA-1069-A2DE-08002B30309D}") != NULL)
                    || (AnsiStrPos(TmpDisplayName2, "{D6277990-4C6A-11CF-8D87-00AA0060F5BF}") != NULL) )
                  {
                    PidlIsGood = false;
                  }
                }  
                else if ((VTVNodeData->NodeData->FNodeLevel == 0) && (VTVNodeData->NodeData->FNodeType == ntNetDrive))
                {
                  //   " - MSN"
                  if (WebMSN.Length() > 0)
                  {                           
                    if (AnsiString(TmpDisplayName2).Pos(WebMSN) != 0)
                    {
                      PidlIsGood = false;
                    }  
                  }
                }
                
                if (PidlIsGood)  
                {
                  if (!DirectoryExists(TmpDisplayName2))
                  {
                    TmpDisplayName2[0] = 0;
                  }

                  ChildNodeStatus  = nsOK;
                  ChildHasChildren = GetHasSubItems(TmpShellFolder, SHCONTF_FOLDERS);
                  ChildNodeType    = GetNewNodeType(NodeType);
                  
                  bool NeedStandartIcon = true;
                  
                  switch (ChildNodeType)
                  {
                    //    
                    case ntDesnaHolderPlant:
                      if (PlantList)
                      {
                        bool Finded = false;
                      
                        for (int k = 0; k < PlantList->Count; k++)
                        {
                          if (PlantList->Names[k] == TmpDisplayName1)
                          {
                            PlantList->Values[PlantList->Names[k]] = "1";
                            Finded = true;
                            break;
                          }
                        }
                        
                        if (!Finded)
                        {
                          ChildNodeStatus = nsWrong;
                        }
                        else
                        {
                          ChildClosedIconIndex   = 1;
                          ChildOpenIconIndex     = 1;
                          NeedStandartIcon       = false;
                        }
                      }
                      else
                      {
                        ChildNodeStatus = nsWrong;
                      }
                      break;

                    //    
                    case ntDesnaHolderUnit:
                      if (UnitList)
                      {
                        bool Finded = false;
                      
                        for (int k = 0; k < UnitList->Count; k++)
                        {
                          if (UnitList->Names[k] == TmpDisplayName1)
                          {
                            UnitList->Values[UnitList->Names[k]] = "1";
                            Finded = true;
                            break;
                          }
                        }
                        
                        if (!Finded)
                        {
                          ChildNodeStatus = nsWrong;
                        }
                        else
                        {
                          ChildClosedIconIndex   = 2;
                          ChildOpenIconIndex     = 2;
                          NeedStandartIcon       = false;
                        }
                      }
                      else
                      {
                        ChildNodeStatus = nsWrong;
                      }
                      break;

                    //    
                    case ntDesnaHolderRepair:
                      if (RepairList)
                      {
                        bool Finded = false;
                      
                        for (int k = 0; k < RepairList->Count; k++)
                        {
                            
                          if (RepairList->Names[k] == TmpDisplayName1)
                          {
                            RepairList->Values[RepairList->Names[k]] = "1";
                            Finded = true;
                            break;
                          }
                        }
                        
                        if (!Finded)
                        {
                          ChildNodeStatus = nsWrong;
                        }
                        else
                        {
                          ChildClosedIconIndex   = 3;
                          ChildOpenIconIndex     = 3;
                          NeedStandartIcon       = false;
                        }
                      }
                      else
                      {
                        ChildNodeStatus = nsWrong;
                      }
                      break;

                    //    
                    case ntDesnaHolderObject:
                      {
                        AnsiString Object = AnsiString(TmpDisplayName1);
                    
                        if (Object == "")
                        {
                          ChildClosedIconIndex   = 4 + 5 * ((FMonitorMode == mmDAT) ? 0 : 1);
                          ChildOpenIconIndex     = ChildClosedIconIndex;
                          NeedStandartIcon       = false;
                        }
                        else if (Object == "")
                        {
                          ChildClosedIconIndex   = 5 + 5 * ((FMonitorMode == mmDAT) ? 0 : 1);
                          ChildOpenIconIndex     = ChildClosedIconIndex;
                          NeedStandartIcon       = false;
                        }
                        else if (Object == "")
                        {
                          ChildClosedIconIndex   = 6 + 5 * ((FMonitorMode == mmDAT) ? 0 : 1);
                          ChildOpenIconIndex     = ChildClosedIconIndex;
                          NeedStandartIcon       = false;
                        }
                        else if (Object == "")
                        {
                          ChildClosedIconIndex   = 7 + 5 * ((FMonitorMode == mmDAT) ? 0 : 1);
                          ChildOpenIconIndex     = ChildClosedIconIndex;
                          NeedStandartIcon       = false;
                        }
                        else if (Object == "")
                        {
                          ChildClosedIconIndex   = 8 + 5 * ((FMonitorMode == mmDAT) ? 0 : 1);
                          ChildOpenIconIndex     = ChildClosedIconIndex;
                          NeedStandartIcon       = false;
                        }
                        else
                        {
                          ChildNodeStatus = nsWrong;
                        }
                      }  
                      break;

                    default:
                      if (!ChildHasChildren)
                      {
                        DWORD Flags = 0;
                        TmpShellFolder->GetAttributesOf(1, (LPCITEMIDLIST *)&TmpPIDL, &Flags);

                        if (Flags & 0x10000) //      ? 
                        {
                          ChildNodeStatus = nsDisabled;
                        }
                        else   //   ?
                        {
                          AnsiString Dir = TmpDisplayName2;

                          int DPIndex = Dir.Pos(":");
                          int SHIndex = Dir.Pos("\\");
                          int Len = Dir.Length();

                          if (DPIndex == Len)
                          {
                            ChildNodeStatus = nsDisabled;
                          }
                          else
                          {
                            if ((DPIndex == (Len - 1)) && (SHIndex == Len))
                            {
                              ChildNodeStatus = nsDisabled;
                            }
                          }
                        }  
                      }
                  }
                  
                  if (VTVNodeData->NodeData->FNodeStatus == nsWrong)
                  {
                    ChildNodeStatus = nsWrong;
                  }
                  
                  if (NeedStandartIcon)
                  {
                    TIcon *DefaultIcon = new TIcon(); 
                    TIcon *OpenIcon = new TIcon(); 

                    //
                    DefaultIcon->Handle = (void *)GetShellImage(TmpFullPIDL, false);

                    //
                    OpenIcon->Handle    = (void *)GetShellImage(TmpFullPIDL, true);

                    ChildClosedIconIndex = AddNewIcon(DefaultIcon);
                    ChildOpenIconIndex    = AddNewIcon(OpenIcon);
                    if (ChildOpenIconIndex < 0)
                    {
                      ChildOpenIconIndex = ChildClosedIconIndex;
                    }
                  
                    DefaultIcon->ReleaseHandle();
                    OpenIcon->ReleaseHandle();
                  
                    delete DefaultIcon;
                    delete OpenIcon;
                  }

                  FCurrentCreatedNodeData = new TNodeData();
                  FCurrentCreatedNodeData->Init(VTVNodeData->NodeData->FNodeLevel + 1, 
                                                ChildNodeType, 
                                                ChildNodeStatus, 
                                                ChildHasChildren, 
                                                ChildClosedIconIndex, 
                                                ChildOpenIconIndex, 
                                                AnsiString(TmpDisplayName2), 
                                                AnsiString(TmpDisplayName1), 
                                                "", 
                                                this, 
                                                TmpPIDL, 
                                                TmpFullPIDL, 
                                                TmpShellFolder);

                  TmpNode = FVTreeViewHolder->AddChild(Node); 

                  VTVChildNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
                  VTVChildNodeData->NodeData = FCurrentCreatedNodeData;

                  if (FCurrentCreatedNodeData->FHasChildren)
                  {
                    TmpNode->States << vsHasChildren;
                  }
                }
                else
                {
                  FpShellMalloc->Free(TmpPIDL);
                  FpShellMalloc->Free(TmpFullPIDL);
                  TmpShellFolder->Release();
                }

                free(reinterpret_cast<void *>(TmpDisplayName1));
                free(reinterpret_cast<void *>(TmpDisplayName2));
              }
            }  
          }
          catch(...)
          {

          }
        }

        switch (NodeType)
        {
          //    
          case ntDesnaHolder:
            if (PlantList)
            {
              for (int k = 0; k < PlantList->Count; k++)
              {
                if (PlantList->Values[PlantList->Names[k]] == "0")
                {
                  ChildNodeType        = ntDesnaHolderPlant; 
                  ChildNodeStatus      = nsDisabled; 
                  ChildHasChildren     = false; 
                  ChildClosedIconIndex = 1; 
                  ChildOpenIconIndex   = 1; 

                  FCurrentCreatedNodeData = new TNodeData();
                  FCurrentCreatedNodeData->Init(VTVNodeData->NodeData->FNodeLevel + 1, 
                                                ChildNodeType, 
                                                ChildNodeStatus, 
                                                ChildHasChildren, 
                                                ChildClosedIconIndex, 
                                                ChildOpenIconIndex, 
                                                "", 
                                                PlantList->Names[k], 
                                                "", 
                                                NULL, 
                                                NULL, 
                                                NULL, 
                                                NULL);

                  TmpNode = FVTreeViewHolder->AddChild(Node); 

                  VTVChildNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
                  VTVChildNodeData->NodeData = FCurrentCreatedNodeData;

                  if (FCurrentCreatedNodeData->FHasChildren)
                  {
                    TmpNode->States << vsHasChildren;
                  }
                }
              } 
            }  
            break;

          //    
          case ntDesnaHolderPlant:
            if (UnitList)
            {
              for (int k = 0; k < UnitList->Count; k++)
              {
                if (UnitList->Values[UnitList->Names[k]] == "0")
                {
                  ChildNodeType        = ntDesnaHolderUnit; 
                  ChildNodeStatus      = nsDisabled; 
                  ChildHasChildren     = false; 
                  ChildClosedIconIndex = 2; 
                  ChildOpenIconIndex   = 2; 

                  FCurrentCreatedNodeData = new TNodeData();
                  FCurrentCreatedNodeData->Init(VTVNodeData->NodeData->FNodeLevel + 1, 
                                                ChildNodeType, 
                                                ChildNodeStatus, 
                                                ChildHasChildren, 
                                                ChildClosedIconIndex, 
                                                ChildOpenIconIndex, 
                                                "", 
                                                UnitList->Names[k], 
                                                "", 
                                                NULL, 
                                                NULL, 
                                                NULL, 
                                                NULL);

                  TmpNode = FVTreeViewHolder->AddChild(Node); 

                  VTVChildNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
                  VTVChildNodeData->NodeData = FCurrentCreatedNodeData;

                  if (FCurrentCreatedNodeData->FHasChildren)
                  {
                    TmpNode->States << vsHasChildren;
                  }
                }
              }
            }
            break;

          //    
          case ntDesnaHolderUnit:  
            if (RepairList)
            {
              for (int k = 0; k < RepairList->Count; k++)
              {
                if (RepairList->Values[RepairList->Names[k]] == "0")
                {
                  ChildNodeType        = ntDesnaHolderRepair; 
                  ChildNodeStatus      = nsDisabled; 
                  ChildHasChildren     = false; 
                  ChildClosedIconIndex = 3; 
                  ChildOpenIconIndex   = 3; 

                  FCurrentCreatedNodeData = new TNodeData();
                  FCurrentCreatedNodeData->Init(VTVNodeData->NodeData->FNodeLevel + 1, 
                                                ChildNodeType, 
                                                ChildNodeStatus, 
                                                ChildHasChildren, 
                                                ChildClosedIconIndex, 
                                                ChildOpenIconIndex, 
                                                "", 
                                                RepairList->Names[k], 
                                                "", 
                                                NULL, 
                                                NULL, 
                                                NULL, 
                                                NULL);

                  TmpNode = FVTreeViewHolder->AddChild(Node); 

                  VTVChildNodeData = (VTVNODEDATA *)FVTreeViewHolder->GetNodeData(TmpNode); 
                  VTVChildNodeData->NodeData = FCurrentCreatedNodeData;

                  if (FCurrentCreatedNodeData->FHasChildren)
                  {
                    TmpNode->States << vsHasChildren;
                  }
                }
              }
            }  
            break;

          //    
          case ntDesnaHolderRepair:
            // ...
            break;

         
          case ntDesnaHolderObject:
            // ...
            break;
        }
        
        if (PlantList)
        {
          delete PlantList;
        }

        if (UnitList)
        {
          delete UnitList;
        }

        if (RepairList)
        {
          delete RepairList;
        }
      }
      __finally
      {
        Iterator->Release();
        Screen->Cursor = crDefault;
      }
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TVTreeHolderOperator::NodeGetImageIndexEvent(TBaseVirtualTree *Sender, PVirtualNode Node, TVTImageKind Kind, 
  TColumnIndex Column, bool &Ghosted, int &ImageIndex)
{
  if (Column == 0)
  {
    switch (Kind)
    {
      case ikNormal: 
      case ikSelected: 
        {
          VTVNODEDATA *VTVNodeData = (VTVNODEDATA *)Sender->GetNodeData(Node); 

          if (VTVNodeData->NodeData)
          {
            if (FHolderImages)
            {
              if (FHolderImages->Count > VTVNodeData->NodeData->FNodeClosedImageIndex)
              {
                if (Sender->Expanded[Node])
                {
                  ImageIndex = VTVNodeData->NodeData->FNodeOpenImageIndex;
                }
                else
                {
                  ImageIndex = VTVNodeData->NodeData->FNodeClosedImageIndex;
                } 
              }
              else
              {
                ImageIndex = -1;
              }
            }
            else
            {
              ImageIndex = -1;
            }
          } 
        }  
        break;
      case ikState: 
      case ikOverlay: 
        ImageIndex = -1;
        break;
    }
  }  
}
//---------------------------------------------------------------------------
void __fastcall TVTreeHolderOperator::NodeGetTextEvent(TBaseVirtualTree *Sender, PVirtualNode Node, 
  TColumnIndex Column, TVSTTextType TextType, WideString &CellText)
{
  if (Column == 0)
  {
    VTVNODEDATA *VTVNodeData = (VTVNODEDATA *)Sender->GetNodeData(Node); 

    if (VTVNodeData->NodeData)
    {
      CellText = VTVNodeData->NodeData->FNodeText;
    }   
    else
    {
      CellText = " ";
    } 
  }  
  else if (Column == 1)
  {
    CellText = "  ";
  } 
  else 
  {
    CellText = "  ...";
  } 
}
//---------------------------------------------------------------------------
void __fastcall TVTreeHolderOperator::NodeBeforePaintEvent( TBaseVirtualTree *Sender, const TCanvas *TargetCanvas,
      PVirtualNode Node, TColumnIndex Column, TVSTTextType TextType)
{
  VTVNODEDATA *VTVNodeData = (VTVNODEDATA *)Sender->GetNodeData(Node); 

  if (VTVNodeData->NodeData)
  {
    if (VTVNodeData->NodeData->FNodeStatus == nsDisabled)
    {                  
      if (!Node->States.Contains(vsSelected))
      {
        TargetCanvas->Font->Color = clGrayText;
      }
      else
      {
        TargetCanvas->Font->Color = clInactiveCaptionText;

        if (!FVTreeViewHolder->Focused())
        {
          TargetCanvas->Brush->Color = clInactiveCaption;
        }
      }
    }  
    else if (VTVNodeData->NodeData->FNodeStatus == nsWrong)
    {                  
      TargetCanvas->Font->Color = clRed;

      if (Node->States.Contains(vsSelected))
      {
        if (!FVTreeViewHolder->Focused())
        {
          TargetCanvas->Brush->Color = clInactiveCaption;
        }
      }
    }  
    else //if (VTVNodeData->NodeData->FNodeStatus == nsOK)
    {  
      if ( (VTVNodeData->NodeData->FNodeType == ntDesnaHolder)
        || (VTVNodeData->NodeData->FNodeType == ntDesnaHolderPlant)
        || (VTVNodeData->NodeData->FNodeType == ntDesnaHolderUnit)
        || (VTVNodeData->NodeData->FNodeType == ntDesnaHolderRepair)
        || (VTVNodeData->NodeData->FNodeType == ntDesnaHolderObject) )
      {
        if (FMonitorMode == mmDAT)
        {
          if (!Node->States.Contains(vsSelected))
          {
            TargetCanvas->Font->Color = clBlue;
          }
          else
          {
            TargetCanvas->Font->Color = clAqua;  

            if (!FVTreeViewHolder->Focused())
            {
              TargetCanvas->Brush->Color = clInactiveCaption;
            }
          }                                         
        }  
        else //if (FMonitorMode == mmRXD)
        {
          if (!Node->States.Contains(vsSelected))
          {
            TargetCanvas->Font->Color = clGreen;
          }
          else
          {
            TargetCanvas->Font->Color = clLime;

            if (!FVTreeViewHolder->Focused())
            {
              TargetCanvas->Brush->Color = clInactiveCaption;
            }
          }
        }
      }
    }
  }     
}
//---------------------------------------------------------------------------
//      STRRET.
//       ,  
//        pMalloc.
//         
//    ::free.
char * __fastcall TVTreeHolderOperator::GetDisplayName(TItemIDList *pidl, const /*STRRET*/TStrRet &Value)
{
  char * result = NULL;
  
  switch( Value.uType )
  {
    case STRRET_CSTR:
      result = strdup( Value.cStr );
      break;

    case STRRET_WSTR:
      {
        int iSize = WideCharToMultiByte( CP_THREAD_ACP, 0, Value.pOleStr, -1, NULL, 0, NULL, NULL );
        result = (char *)malloc(iSize);
        WideCharToMultiByte( CP_THREAD_ACP, 0, Value.pOleStr, -1, result, iSize, NULL, NULL );
        FpShellMalloc->Free( Value.pOleStr );
      }
      break;
      
    case STRRET_OFFSET:
      result = strdup( PSTR( PBYTE(pidl) + Value.uOffset ) );
      break;
  }
  return result;
}
//---------------------------------------------------------------------------
void * __fastcall TVTreeHolderOperator::GetShellImage(PItemIDList PIDL, bool Open)
{
  TSHFileInfo FileInfo;
    
  int Flags = SHGFI_PIDL | SHGFI_SHELLICONSIZE |SHGFI_ICON | SHGFI_SMALLICON;

  if(Open)
  {
    Flags |= SHGFI_OPENICON;
  }    

  Flags |= SHGFI_SMALLICON;

  SHGetFileInfo(PChar(PIDL), 0, &FileInfo, sizeof(FileInfo), Flags);

  return FileInfo.hIcon;
}
//---------------------------------------------------------------------------
bool __fastcall TVTreeHolderOperator::GetHasSubItems(_di_IShellFolder ShellFolder, int Flags)
{
    PItemIDList ID;
    IEnumIDList *EnumList;
    DWORD NumIDs;
    HResult HR;
    int ErrMode;
    bool Result = False;

    if(ShellFolder)
    {
      ErrMode = SetErrorMode(SEM_FAILCRITICALERRORS);

      try
      {
        HR = ShellFolder->EnumObjects(0, Flags, &EnumList);
        if(HR == S_OK)
        {
          Result = EnumList->Next(1, &ID, &NumIDs) == S_OK;
        }  
      }
      __finally
      {
        SetErrorMode(ErrMode);
      }
    }
      
    return Result;
}
//---------------------------------------------------------------------------
PItemIDList __fastcall TVTreeHolderOperator::ConcatPIDLs(PItemIDList IDList1, PItemIDList IDList2)
{
  int cb1, cb2;

  if(IDList1 != NULL)
  {
    cb1 = GetPIDLSize(IDList1) - sizeof(IDList1->mkid.cb);
  }  
  else
  {
    cb1 = 0;
  }  

  cb2 = GetPIDLSize(IDList2);

  PItemIDList PIDL = CreatePIDL(cb1 + cb2);
  
  if(PIDL != NULL)
  {
    if(IDList1 != NULL)
    {
      memcpy(PIDL, IDList1, cb1);
    }  

    memcpy((PChar)PIDL + cb1, IDList2, cb2);
  }

  return PIDL;
}
//---------------------------------------------------------------------------
NT __fastcall TVTreeHolderOperator::GetNewNodeType(NT ANodeType)
{
  switch (ANodeType)
  {
    case ntNone:
      return ANodeType;

    case ntLocalDrive:
      return ANodeType;

    case ntNetDrive:
      return ANodeType;

    case ntFavorite:
      return ANodeType;

    case ntDesnaHolder:
      return ntDesnaHolderPlant;

    case ntDesnaHolderPlant:
      return ntDesnaHolderUnit;

    case ntDesnaHolderUnit:
      return ntDesnaHolderRepair;

    case ntDesnaHolderRepair:
      return ntDesnaHolderObject;

    case ntDesnaHolderObject:
      return ntNone;

    default:  
      return ntNone;
  }
}
//---------------------------------------------------------------------------
int __fastcall TVTreeHolderOperator::AddNewIcon(TIcon *AIcon)
{
  int Result = -1;

  if (FHolderImages && AIcon)
  {
    int Index = FHolderImages->AddIcon(AIcon);
    
    if (Index > -1)
    {
      Graphics::TBitmap *IconBitmap1 = new Graphics::TBitmap();

      FHolderImages->GetBitmap(Index, IconBitmap1);
  
      for (int k = FImageListConstCount; k < (FHolderImages->Count - 1); k++)
      {
        Graphics::TBitmap *IconBitmap2 = new Graphics::TBitmap();   
        FHolderImages->GetBitmap(k, IconBitmap2);

        int y = 0;
        int x = 0;
      
        for (; y < IconBitmap1->Height; y++)
        {
          x = 0;
        
          for (; x < IconBitmap1->Width; x++)
          {
            if (IconBitmap1->Canvas->Pixels[x][y] != IconBitmap2->Canvas->Pixels[x][y])
            {
              break;
            }
          }

          if (x < 16)
          {
            break;
          }
        }

        if ((x > 15) && (y > 15))
        {
          Result = k;
          FHolderImages->Delete(Index);
        }

        delete IconBitmap2;  
      }

      delete IconBitmap1;  
      
      if (Result < 0)
      {
        Result = Index;
      }
    }  
  }

  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TVTreeHolderOperator::GetPIDLSize(PItemIDList IDList)
{
  int PIDLSize = 0;
  if(IDList)
  {
    PIDLSize = sizeof(IDList->mkid.cb);
    while(IDList->mkid.cb)
    {
      PIDLSize += IDList->mkid.cb;
      IDList = NextPIDL(IDList);
    }
  }
  return PIDLSize;
}
//---------------------------------------------------------------------------
PItemIDList __fastcall TVTreeHolderOperator::CreatePIDL(int Size)
{
  PItemIDList PIDL = NULL;

  if(FpShellMalloc)
  {  
    PIDL = (PItemIDList)FpShellMalloc->Alloc(Size);

    if(PIDL != NULL)
    {
      memset(PIDL, 0, Size);
    }
  }  
  
  return PIDL;
}
//---------------------------------------------------------------------------
PItemIDList __fastcall TVTreeHolderOperator::NextPIDL(PItemIDList IDList)
{
  return (PItemIDList)((int)IDList+IDList->mkid.cb); 
}
//---------------------------------------------------------------------------
#pragma package(smart_init)


